// const { resolve, sep } = require('path')
// const fs = require('fs')
import Database from 'better-sqlite3';

const serialize = require('serialize-javascript');

const {
  get: _get,
  set: _set,
  has: _has,
  isNil,
  // isFunction,
  // isArray,
  // isObject,
  // toPath,
  // merge,
  // clone,
  // cloneDeep,
} = require('lodash');

interface EnmapOptions<V, SV> {
  name ?: string
  dataDir ?: string
  verbose ?: boolean
  wal ?: boolean
  serializer ?: (value: V, key: string) => SV;
  deserializer ?: (value: SV, key: string) => V;
}

export default class Enmap<
  V = any,
  SV = unknown,
> {
  // derived from user options
  // #verbose
  #wal
  #name
  #serializer
  #deserializer

  // internal references only
  #db

  constructor(options : EnmapOptions<V, SV>) {
    // this.#verbose = options.verbose ? options.verbose : () => null
    this.#wal = options.wal ?? true
    this.#name = this.#parseName(options.name)
    this.#serializer = options.serializer ? options.serializer : (data: V): V => data;
    this.#deserializer = options.deserializer
      ? options.deserializer
      : (data: V): V => data;

    this.#db = new Database(':memory:') // { verbose: this.#verbose } ?

    const table = this.#db
      .prepare(
        "SELECT count(*) FROM sqlite_master WHERE type='table' AND name = ?;",
      )
      .get(this.#name);
    if (!table['count(*)']) {
      this.#db
        .prepare(
          `CREATE TABLE ${this.#name} (key text PRIMARY KEY, value text)`,
        )
        .run();
      this.#db.pragma('synchronous = 1');
      if (this.#wal) this.#db.pragma('journal_mode = wal');
    }
    this.#db
      .prepare(
        `CREATE TABLE IF NOT EXISTS 'internal::autonum' (enmap TEXT PRIMARY KEY, lastnum INTEGER)`,
      )
      .run();
  }

  // ALL GETTERS
  get name(): string {
    return this.#name
  }

  get(key : string) : V | null {
    if (isNil(key)) return null;
    const data = this.#db
      .prepare(`SELECT * FROM ${this.#name} WHERE key = ?;`)
      .get(key);
    if (!data) return null;
    return this.#parseData(data.key, data.value)
  }

  get keys(): string[] {
    const keys = this.#db.prepare(`SELECT key FROM ${this.#name}`).all()
    return keys.map(res => res.key);
  }

  get values(): V[] {
    const values = this.#db.prepare(`SELECT value FROM ${this.#name}`).all()
    return values.map(res => this.#parseData(res.key, res.value));
  }

  forEach(callback: (value: V, index: string, array: any[]) => void): void {
    const data = this.#db.prepare(`SELECT key, value FROM ${this.#name}`).all()
    data.forEach(({ key, value }) => callback(this.#parseData(key, value), key, data))
  }

  map(callback: (value: V, index: string, array: any[]) => any): any[] {
    const data = this.#db.prepare(`SELECT key, value FROM ${this.#name}`).all()
    return data.map(({ key, value }) => callback(this.#parseData(key, value), key, data))
  }

  get size() {
    const data = this.#db
      .prepare(`SELECT count(*) FROM '${this.#name}';`)
      .get();
    return data['count(*)'];
  }

  // ALL SETTERS

  set(key : string, value : V): this | null {
    if (isNil(key)) return null;
    const serialized = serialize(this.#serializer(value, key));
    this.#db
      .prepare(
        `INSERT OR REPLACE INTO ${this.#name} (key, value) VALUES (?, ?);`,
      )
      .run(key, serialized);
    return this
  }

  #parseData(key : string, value : string) : V {
    return this.#deserializer(eval(`(${value})`), key);
  }

  #parseName(name? : string) : string {
    return name?.toLowerCase().replace(/[\W_]+/g, '-') ?? 'MemoryEnmap'
  }

}

